/*
 *  $Id: context_util.c,v 1.4 1993/11/12 12:31:42 doss Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

/* context_util.c - Utilities used by the context chapter functions */

#include <mpi.h>


#define MPIR_MAX(a,b)  (((a)>(b))?(a):(b))

/*+

MPIR_Context_alloc - allocate some number of contexts over given communicator

+*/
int MPIR_Context_alloc ( comm, num_contexts, context )
MPI_Comm comm;
int num_contexts;
uint32 *context;
{
  static uint32 high_context = MPIR_FIRST_FREE_CONTEXT;
  int n, N, N2_prev, surfeit;
  int d, dst, src;
  int context_in;

  /* Intialize communicator size */
  (void) MPI_Comm_size ( comm, &N );
  
  /* If there's only one member, this is trivial */
  if ( N > 1 ) {

    /* Initialize variables */
    MPI_Comm_request s_req, r_req;
    MPI_Status status;
    MPI_Comm comm_coll = comm->comm_coll;
    (void) MPI_Comm_rank ( comm, &n );
    N2_prev = comm->group->N2_prev;
    surfeit = N - N2_prev;

    /* Perform a combine operation to find maximum unused context */
    if ( n < N2_prev ) {
      if( n < surfeit ) {

		/* get the fanin letter from the upper "half" process: */
		dst = N2_prev + n;

        MPI_Irecv( &context_in, 1, MPI_INT, dst, 0, comm_coll, &r_req );
        MPI_Wait( r_req, &status );

		/* determine max context */
        high_context = MPIR_MAX( high_context, context_in );
      }

      /* combine on embedded N2_prev power-of-two processes */
      for (d = 1; d < N2_prev; d <<= 1) {
        dst = (n ^ d);

        /* Send and receive a context limit */
        if ( n < dst ) {
          MPI_Isend( &high_context, 1, MPI_INT, dst, 0, comm_coll, &s_req );
          MPI_Wait( s_req, &status );
          MPI_Irecv( &context_in, 1, MPI_INT, dst, 0, comm_coll, &r_req );
          MPI_Wait( r_req, &status );
        }
        else {
          MPI_Irecv( &context_in, 1, MPI_INT, dst, 0, comm_coll, &r_req );
          MPI_Wait( r_req, &status );
          MPI_Isend( &high_context, 1, MPI_INT, dst, 0, comm_coll, &s_req );
          MPI_Wait( s_req, &status );
        }

		/* determine max context */
        high_context = MPIR_MAX( high_context, context_in );
      }

      /* fanout data to nodes above N2_prev... */
      if ( n < surfeit ) {
		dst = N2_prev + n;
        MPI_Isend( &high_context, 1, MPI_INT, dst, 0, comm_coll, &s_req );
        MPI_Wait( s_req, &status );
      }
	} 
	else {
	    /* fanin data to power of 2 subset */
	    src = n - N2_prev;
        MPI_Isend( &high_context, 1, MPI_INT, src, 0, comm_coll, &s_req );
        MPI_Wait( s_req, &status );
        MPI_Irecv( &high_context, 1, MPI_INT, src, 0, comm_coll, &r_req );
        MPI_Wait( r_req, &status );
	}
  } 
  
  /* Set the context and reset the highest context */
  (*context) = high_context;
  high_context += num_contexts;

  return(MPI_SUCCESS); 
}

/*+

MPIR_Context_dealloc - deallocate previously allocated contexts 

+*/
/*ARGSUSED*/
int MPIR_Context_dealloc ( comm, num, context )
MPI_Comm comm;
int num;
uint32 context;
{
  /* This does nothing currently */
  return (MPI_SUCCESS);
}
